/* ------------------------------------------------------------------------ */
/* matrix.h - Funktionsbibliothek zur Arbeit mit Vektoren und Matrizen      */
/*                                                                          */
/*               Copyright Tilman Kuepper 2009 - 2020.                      */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 5. Juli 2020                                         */
/* ------------------------------------------------------------------------ */

#ifndef MATRIX_H_Included
#define MATRIX_H_Included

#include <stdio.h>  /* fuer FILE und size_t */

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------------------------------------------------------ */
/* Rueckgabewerte bzw. Fehlermeldungen von Funktionen.                      */
/* ------------------------------------------------------------------------ */
#define MAT_OK         0
#define MAT_ERR_NOINV  1
#define MAT_ERR_FOPEN  2
#define MAT_ERR_FWRITE 3
#define MAT_ERR_FORMAT 4
#define MAT_ERR_MEMORY 5
#define MAT_ERR_NOTEQ  6
#define MAT_ERR_DIM    7
#define MAT_ERR_VNORM  8
#define MAT_ERR_ITFAIL 9

/* ------------------------------------------------------------------------ */
/* Rueckgabewerte bzw. Fehlermeldungen in Textform.                         */
/* Fuer den genauen Text der Meldungen: siehe Datei "matrix.c".             */
/* ------------------------------------------------------------------------ */
extern const char *mat_errmsg[];

/* ------------------------------------------------------------------------ */
/* Parameter zur Uebergabe an die Funktionen "m_init" und "v_init".         */
/* ------------------------------------------------------------------------ */
#define MAT_INIT_NULL        0     /* Alle Elemente auf null setzen         */
#define MAT_INIT_RANDOM      1     /* Zufallszahlen zwischen 0...1 erzeugen */
#define MAT_INIT_IDENTITY    2     /* Matrix auf Einheitsmatrix setzen      */
#define MAT_INIT_HILBERT     3     /* Matrix auf Hilbertmatrix setzen       */
#define MAT_INIT_INV_HILBERT 4     /* Matrix auf inv. Hilbertmatrix setzen  */
#define MAT_INIT_ONES        5     /* Alle Elemente auf eins setzen         */

/* ------------------------------------------------------------------------ */
/* Parameter zur Uebergabe an die Funktion "m_make_symmetric".              */
/* ------------------------------------------------------------------------ */
#define MAT_USE_MEAN  0    /* Oberes und unteres Dreieck der Matrix mitteln */
#define MAT_USE_UPPER 1    /* Oberes Dreieck der Matrix nach unten kopieren */
#define MAT_USE_LOWER 2    /* Unteres Dreieck der Matrix nach oben kopieren */

/* ------------------------------------------------------------------------ */
/* Folgende Makros werden zur numerischen Nullstellenberechnung benoetigt.  */
/* ------------------------------------------------------------------------ */

/* Gewuenschte Genauigkeit */
#ifndef MAT_EPSILON
#define MAT_EPSILON 1.0e-10
#endif

/* Maximale Anzahl von Iterationen des Loesungsverfahrens */
#ifndef MAT_MAX_ITER
#define MAT_MAX_ITER 1000000UL
#endif

/* Komplexer Startwert fuer das WDK-Verfahren zur Nullstellenbestimmung */
#ifndef MAT_WDK_INIT
#define MAT_WDK_INIT 0.4, 0.9  /* 0.4 + 0.9 i */
#endif

/* ------------------------------------------------------------------------ */
/* Datentyp fuer komplexe Zahlen. Dieser wird bei der Verarbeitung von      */
/* Polynomen - zum Beispiel zur Nullstellenbestimmung - benoetigt.          */
/* ------------------------------------------------------------------------ */
typedef struct tagHM_COMPLEX
{
    double val[2];  /* val[0] ist Realteil, val[1] ist Imaginaerteil */
} HM_COMPLEX;

/* Zugriff auf Real- und Imaginaerteil */
#undef MAT_REAL
#undef MAT_IMAG

#define MAT_REAL(z) z.val[0]
#define MAT_IMAG(z) z.val[1]

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird initialisiert (je nach Parameter "type").          */
/* ------------------------------------------------------------------------ */
void m_init(size_t dim, void *mat, int type);

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Matrix "mat" werden auf den angegebenen Wert gesetzt.  */
/* ------------------------------------------------------------------------ */
void m_fill(size_t dim, void *mat, double value);

/* ------------------------------------------------------------------------ */
/* Die Elemente der Matrix "mat" werden auf dem Bildschirm ausgegeben.      */
/* Das Format kann ueber die Konstante MAT_PRN_FMT eingestellt werden.      */
/* ------------------------------------------------------------------------ */
void m_print(size_t dim, const void *mat);

/* ------------------------------------------------------------------------ */
/* Der Inhalt der Matrix "mat" wird nach "mresult" kopiert.                 */
/* ------------------------------------------------------------------------ */
void m_copy(size_t dim, void *result, const void *mat);

/* ------------------------------------------------------------------------ */
/* Die Matrizen "mat1" und "mat2" werden verglichen. Falls beide Matrizen   */
/* identisch sind, wird MAT_OK zurueckgegeben, andernfalls MAT_ERR_NOTEQ.   */
/* ------------------------------------------------------------------------ */
int m_equal(size_t dim, const void *mat1, const void *mat2);

/* ------------------------------------------------------------------------ */
/* Die Matrizen "mat1" und "mat2" werden multipliziert.                     */
/* Das Ergebnis wird in der Matrix "mresult" abgelegt.                      */
/* ------------------------------------------------------------------------ */
void m_mul(size_t dim, void *result, const void *mat1, const void *mat2);

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Matrix "mat" werden mit dem Faktor "f" multipliziert.  */
/* ------------------------------------------------------------------------ */
void m_fmul(size_t dim, void *mat, double f);

/* ------------------------------------------------------------------------ */
/* Die Matrizen "mat1" und "mat2" werden addiert. Das Ergebnis              */
/* wird in der Matrix "mresult" abgelegt.                                   */
/* ------------------------------------------------------------------------ */
void m_add(size_t dim, void *result, const void *mat1, const void *mat2);

/* ------------------------------------------------------------------------ */
/* Das "f-fache" der Matrix "mat" wird zur Matrix "mresult" addiert.        */
/* ------------------------------------------------------------------------ */
void m_fadd(size_t dim, void *result, const void *mat, double f);

/* ------------------------------------------------------------------------ */
/* Die Zeilen "row1" und "row2" der Matrix "mat" werden vertauscht.         */
/* ------------------------------------------------------------------------ */
void m_swap_row(size_t dim, void *mat, size_t row1, size_t row2);

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Zeile "row" werden mit einem Faktor multipliziert.     */
/* ------------------------------------------------------------------------ */
void m_mul_row(size_t dim, void *mat, size_t row, double factor);

/* ------------------------------------------------------------------------ */
/* Alle Elemente der Spalte "col" werden mit einem Faktor multipliziert.    */
/* ------------------------------------------------------------------------ */
void m_mul_col(size_t dim, void *mat, size_t col, double factor);

/* ------------------------------------------------------------------------ */
/* Von der Zeile "result" wird das "f-fache" der Zeile "sub" abgezogen.     */ 
/* ------------------------------------------------------------------------ */
void m_sub_row(size_t dim, void *mat, size_t result, size_t sub, double f);

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird invertiert. Falls dabei ein Fehler auftritt, ist   */
/* der Rueckgabewert MAT_ERR_MEMORY (nicht genuegend freier Speicher) oder  */
/* MAT_ERR_NOINV (inverse Matrix existiert nicht). Andernfalls wird die     */
/* inverse Matrix in "mat" abgelegt und der Rueckgabewert ist MAT_OK.       */
/* ------------------------------------------------------------------------ */
int m_invert(size_t dim, void *mat);

/* ------------------------------------------------------------------------ */
/* Die Determinante der Matrix "mat" wird berechnet und zurueckgegeben.     */
/* Falls nicht genuegend freier Speicher zur Verfuegung steht, wird als     */
/* Ergebnis 0 zurueckgegeben und *pResult auf MAT_ERR_MEMORY gesetzt,       */
/* andernfalls auf MAT_OK. Es ist moeglich, fuer pResult NULL zu            */
/* uebergeben, falls keine Erfolgs-/Fehlermeldung erforderlich ist.         */
/* ------------------------------------------------------------------------ */
double m_det(size_t dim, const void *mat, int *pResult);

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird an der Hauptdiagonalen gespiegelt (transponiert).  */
/* ------------------------------------------------------------------------ */
void m_transpose(size_t dim, void *mat);

/* ------------------------------------------------------------------------ */
/* Ist die Matrix "mat" symmetrisch (Rueckgabe = MAT_OK) oder ist sie nicht */
/* symmetrisch (Rueckgabe = MAT_ERR_NOTEQ)? Bei einer symmetrischen Matrix  */
/* weichen die an der Hauptdiagonalen gespiegelten Matrixelemente maximal   */
/* um den Betrag "epsilon" voneinander ab.                                  */
/* ------------------------------------------------------------------------ */
int m_is_symmetric(size_t dim, const void *mat, double epsilon);

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird symmetrisch gemacht. Entweder dadurch, dass        */
/* die Elemente oberhalb der Hauptdiagonalen nach unten kopiert werden      */
/* (mode = MAT_USE_UPPER) oder dadurch, dass die Elemente unterhalb der     */
/* Hauptdiagonalen nach oben kopiert werden (mode = MAT_USE_LOWER) oder     */
/* dadurch, dass die Mittelwerte aller an der Hauptdiagonalen gespiegelten  */
/* Elemente berechnet werden (mode = MAT_USE_MEAN).                         */
/* ------------------------------------------------------------------------ */
void m_make_symmetric(size_t dim, void *mat, int mode);

/* ------------------------------------------------------------------------ */
/* Es wird die QR-Zerlegung der Matrix "mat" ermittelt. Der Rueckgabewert   */
/* ist entweder MAT_ERR_MEMORY (zu wenig freier Speicher) oder MAT_OK.      */
/* ------------------------------------------------------------------------ */
int m_qr_decomp(size_t dim, const void *mat, void *q, void *r);

/* ------------------------------------------------------------------------ */
/* Es wird die LU-Zerlegung P * MAT = L * U der Matrix "mat" ermittelt.     */
/* Die Ergebnismatrix "lu" enthaelt im oberen Dreieck die Matrix U und      */
/* unterhalb der Hauptdiagonalen L. Die Diagonalelemente von L sind nicht   */
/* in "lu" abgelegt; sie sind alle gleich eins. Die waehrend der Zerlegung  */
/* durchgefuehrten Zeilenvertauschungen sind in der Permutationsmatrix "p"  */
/* wiedergegeben. Mittels "sgn" wird der Wert (-1)^n zurueckgegeben,        */
/* wobei "n" die Anzahl der durchgefuehrten Zeilenvertauschungen ist.       */
/* Die Rueckgabe ist MAT_OK bzw. MAT_ERR_NOINV (wenn "mat" singulaer ist).  */
/* ------------------------------------------------------------------------ */
int m_lu_decomp(size_t dim, const void *mat, void *lu, void *p, double *sgn);

/* ------------------------------------------------------------------------ */
/* Die Koeffizienten des charakteristischen Polynoms der Matrix "mat"       */
/* werden ermittelt und in den Vektor "pol" geschrieben. Im Parameter "dim" */
/* ist die Dimension der Matrix "mat" zu uebergeben. Hinweis: Es werden     */
/* insgesamt (dim + 1) Koeffizienten ermittelt und in "pol" abgelegt.       */
/* Der Ergebnisvektor "pol" muss daher (dim + 1) Elemente besitzen. Die     */
/* Rueckgabe ist MAT_OK oder MAT_ERR_MEMORY (zu wenig freier Speicher).     */
/* ------------------------------------------------------------------------ */
int m_charpol(size_t dim, const void *mat, double *pol);

/* ------------------------------------------------------------------------ */
/* Mittels zyklischem Jacobi-Verfahren werden alle Eigenwerte und -vektoren */
/* der symmetrischen Matrix "mat" ermittelt. Die Eigenwerte werden auf der  */
/* Hauptdiagonalen der Matrix "eig", die Eigenvektoren in den jeweiligen    */
/* Spalten der Matrix "vec" abgelegt. Die Eigenwerte sind nicht nach ihrer  */
/* Groesse sortiert. Die Rueckgabe der Funktion "m_jacobi" ist entweder     */
/* MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL (wenn das Jacobi-Verfahren    */
/* nach "MAT_MAX_ITER" Iterationen kein Ergebnis mit der Genauigkeit        */
/* "MAT_EPSILON" ermitteln konnte).                                         */
/*                                                                          */  
/* Achtung: Die Funktion "m_jacobi" darf nur fuer symmetrische Matrizen     */
/* "mat" aufgerufen werden. Wenn die Matrix "mat" nicht symmetrisch ist,    */
/* werden keine korrekten Ergebnisse berechnet!                             */
/* ------------------------------------------------------------------------ */
int m_jacobi(size_t dim, const void *mat, void *eig, void *vec);

/* ------------------------------------------------------------------------ */
/* Mittels zyklischem Jacobi-Verfahren werden alle Eigenwerte und -vektoren */
/* der symmetrischen Matrix "mat" ermittelt. Die Eigenwerte werden auf der  */
/* Hauptdiagonalen der Matrix "eig", die Eigenvektoren in den jeweiligen    */
/* Spalten der Matrix "vec" abgelegt. Die Eigenwerte sind nicht nach ihrer  */
/* Groesse sortiert. Die Rueckgabe der Funktion "m_jacobi2" ist entweder    */
/* MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL (wenn das Jacobi-Verfahren    */
/* innerhalb von "max_iter" Iterationen kein Ergebnis mit der Genauigkeit   */
/* "epsilon" ermitteln konnte).                                             */
/*                                                                          */  
/* Achtung: Die Funktion "m_jacobi2" darf nur fuer symmetrische Matrizen    */
/* "mat" aufgerufen werden. Wenn die Matrix "mat" nicht symmetrisch ist,    */
/* werden keine korrekten Ergebnisse berechnet!                             */
/* ------------------------------------------------------------------------ */
int m_jacobi2(size_t dim, const void *mat, void *eig, void *vec,
    double epsilon, size_t max_iter);

/* ------------------------------------------------------------------------ */
/* Der betragsmaessig groesste Eigenwert der Matrix "mat" und der dazu      */
/* passende Eigenvektor werden ermittelt und ueber die (Zeiger-)Parameter   */
/* "eig" und "vec" an den Aufrufer zurueckgegeben. Der Rueckgabewert der    */
/* Funktion "m_mises2" betraegt MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL  */
/* (falls nach "max_iter" Iterationen kein Ergebnis mit der Genauigkeit     */
/* "epsilon" ermittelt werden konnte).                                      */
/*                                                                          */
/* Achtung: Die Berechnung des Eigenwerts bzw. des Eigenvektors erfolgt     */
/* ueber eine von-Mises-Vektoriteration. Insbesondere bei doppelten, eng    */
/* benachbarten oder komplexen Eigenwerten konvergiert das Verfahren nur    */
/* langsam oder ueberhaupt nicht. In solchen Faellen sollten die Parameter  */
/* "max_iter" und "epsilon" auf geeignete Werte gesetzt werden, um ein      */
/* Einfrieren des Programmablaufs zu vermeiden.                             */
/* ------------------------------------------------------------------------ */
int m_mises2(size_t dim, const void *mat, double *eig, double *vec,
    size_t max_iter, double epsilon);

/* ------------------------------------------------------------------------ */
/* Der betragsmaessig groesste Eigenwert der Matrix "mat" und der dazu      */
/* passende Eigenvektor werden ermittelt und ueber die (Zeiger-)Parameter   */
/* "eig" und "vec" an den Aufrufer zurueckgegeben. Der Rueckgabewert der    */
/* Funktion "m_mises" betraegt MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_ITFAIL   */
/* (falls nach "MAT_MAX_ITER" Iterationen kein Ergebnis mit der Genauigkeit */
/* "MAT_EPSILON" ermittelt werden konnte).                                  */
/*                                                                          */
/* Achtung: Die Berechnung des Eigenwerts bzw. des Eigenvektors erfolgt     */
/* ueber eine von-Mises-Vektoriteration. Insbesondere bei doppelten, eng    */
/* benachbarten oder komplexen Eigenwerten konvergiert das Verfahren nur    */
/* langsam oder ueberhaupt nicht. In diesen Faellen ist es sinnvoll, statt  */
/* "m_mises" die Funktion "m_mises2" aufzurufen und dort das Iterationsende */
/* ueber die Parameter "max_iter" bzw. "epsilon" gezielt einzustellen.      */
/* ------------------------------------------------------------------------ */
int m_mises(size_t dim, const void *mat, double *eig, double *vec);

/* ------------------------------------------------------------------------ */
/* Einen einzelnen Zeilenvektor der Matrix "mat" zurueckgeben.              */
/* ------------------------------------------------------------------------ */
void m_get_row(size_t dim, size_t row, const void *mat, double *vec);

/* ------------------------------------------------------------------------ */
/* Einen einzelnen Spaltenvektor der Matrix "mat" zurueckgeben.             */
/* ------------------------------------------------------------------------ */
void m_get_col(size_t dim, size_t col, const void *mat, double *vec);

/* ------------------------------------------------------------------------ */
/* Die Hauptdiagonalelemente von "mat" in Form eines Vektors zurueckgeben.  */
/* ------------------------------------------------------------------------ */
void m_get_diag(size_t dim, const void *mat, double *vec);

/* ------------------------------------------------------------------------ */
/* Das maximale Element der Matrix "mat" wird zurueckgegeben.               */
/* ------------------------------------------------------------------------ */
double m_get_max(size_t dim, const void *mat);

/* ------------------------------------------------------------------------ */
/* Das minimale Element der Matrix "mat" wird zurueckgegeben.               */
/* ------------------------------------------------------------------------ */
double m_get_min(size_t dim, const void *mat);

/* ------------------------------------------------------------------------ */
/* Der Vektor "vec" wird initialisiert (je nach Parameter "type").          */
/* ------------------------------------------------------------------------ */
void v_init(size_t dim, double *vec, int type);

/* ------------------------------------------------------------------------ */
/* Alle Elemente des Vektors "vec" werden auf den angegebenen Wert gesetzt. */
/* ------------------------------------------------------------------------ */
void v_fill(size_t dim, double *vec, double value);

/* ------------------------------------------------------------------------ */
/* Alle Elemente des Vektors "src" werden in den Vektor "vec" kopiert.      */
/* ------------------------------------------------------------------------ */
void v_copy(size_t dim, double *vec , const double *src);

/* ------------------------------------------------------------------------ */
/* Die Elemente des Vektors "vec" werden auf dem Bildschirm ausgegeben.     */
/* Das Format kann ueber die Konstante MAT_PRN_FMT eingestellt werden.      */
/* ------------------------------------------------------------------------ */
void v_print(size_t dim, const double *vec);

/* ------------------------------------------------------------------------ */
/* Die Vektoren "v1" und "v2" werden verglichen. Falls beide Vektoren       */
/* identisch sind, wird MAT_OK zurueckgegeben, andernfalls MAT_ERR_NOTEQ.   */
/* ------------------------------------------------------------------------ */
int v_equal(size_t dim, const double *v1, const double *v2);

/* ------------------------------------------------------------------------ */
/* Der Betrag des Vektors "v1" wird berechnet.                              */
/* ------------------------------------------------------------------------ */
double v_abs(size_t dim, const double *v1);

/* ------------------------------------------------------------------------ */
/* Der Vektor "v1" wird normalisiert (Rueckgabe == MAT_OK). Hat "v1" die    */
/* Laenge null, bleibt "v1" unveraendert (Rueckgabe == MAT_ERR_VNORM).      */
/* ------------------------------------------------------------------------ */
int v_normalize(size_t dim, double *v1);

/* ------------------------------------------------------------------------ */
/* Das "f-fache" des Vektors "v1" wird zum Vektor "vresult" addiert.        */
/* ------------------------------------------------------------------------ */
void v_fadd(size_t dim, double *vresult, const double *v1, double f);

/* ------------------------------------------------------------------------ */
/* Der Vektor "vresult" wird mit "factor" multipliziert.                    */
/* ------------------------------------------------------------------------ */
void v_fmul(size_t dim, double *vresult, double factor);

/* ------------------------------------------------------------------------ */
/* Das Skalarprodukt der Vektoren "v1" und "v2" wird berechnet.             */
/* ------------------------------------------------------------------------ */
double v_dot(size_t dim, const double *v1, const double *v2);

/* ------------------------------------------------------------------------ */
/* Das Kreuzprodukt der dreidimensionalen (!) Vektoren "v1" und "v2" wird   */
/* berechnet und im Vektor "vresult" abgespeichert.                         */
/* ------------------------------------------------------------------------ */
void v_cross(double *vresult, const double *v1, const double *v2);

/* ------------------------------------------------------------------------ */
/* Das maximale Element des Vektors "vec" wird zurueckgegeben.              */
/* ------------------------------------------------------------------------ */
double v_get_max(size_t dim, const double *vec);

/* ------------------------------------------------------------------------ */
/* Das minimale Element des Vektors "vec" wird zurueckgegeben.              */
/* ------------------------------------------------------------------------ */
double v_get_min(size_t dim, const double *vec);

/* ------------------------------------------------------------------------ */
/* Das Produkt aus der Matrix "mat" und dem Vektor "v" wird berechnet und   */
/* im Vektor "vresult" abgelegt.                                            */
/* ------------------------------------------------------------------------ */
void m_vmul(size_t dim, double *vresult, const void *mat, const double *v);

/* ------------------------------------------------------------------------ */
/* Der Wert des Polynoms a[0] x^n + a[1] x^(n-1) + a[2] x^(n-2)... + a[n]   */
/* wird fuer den angegebenen komplexen Wert "x" berechnet. Die Anzahl der   */
/* komplexen Koeffizienten ist im Parameter "dim" zu uebergeben.            */
/* ------------------------------------------------------------------------ */
HM_COMPLEX v_cpolyval(size_t dim, const HM_COMPLEX *a, HM_COMPLEX x);

/* ------------------------------------------------------------------------ */
/* Der Wert des Polynoms a[0] x^n + a[1] x^(n-1) + a[2] x^(n-2)... + a[n]   */
/* wird fuer den angegebenen reellen Wert "x" berechnet. Die Anzahl der     */
/* reellen Koeffizienten ist im Parameter "dim" zu uebergeben.              */
/* ------------------------------------------------------------------------ */
double v_dpolyval(size_t dim, const double *a, double x);

/* ------------------------------------------------------------------------ */
/* Alle reellen und komplexen Nullstellen eines Polynoms werden ermittelt   */
/* und im Vektor "roots" abgelegt. Die komplexen Koeffizienten des Polynoms */
/* sind in a[0], ..., a[n] zu uebergeben, deren Anzahl ist im Parameter     */
/* "dim" zu uebergeben. Hinweis: Ein Polynom mit z Koeffizienten besitzt    */
/* genau (z-1) Nullstellen. Der Ergebnisvektor "roots" hat also ein Element */
/* weniger als Vektor "a". Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (Algorithmus   */
/* konvergiert nicht) oder MAT_ERR_DIM (falls dim < 2 oder a[0] = 0).       */
/* ------------------------------------------------------------------------ */
int v_croots(size_t dim, const HM_COMPLEX *a, HM_COMPLEX *roots);

/* ------------------------------------------------------------------------ */
/* Siehe v_croots(). Bei dieser Variante kann zusaetzlich die gewuenschte   */
/* Genauigkeit, die maximale Anzahl der Iterationen sowie der Startwert     */
/* fuer das verwendete Weierstrass-Durand-Kerner-Iterationsverfahren        */
/* angegeben werden. Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (der Algorithmus     */
/* konvergiert nicht) oder MAT_ERR_DIM (falls dim < 2 oder a[0] = 0).       */
/* ------------------------------------------------------------------------ */
int v_croots2(size_t dim, const HM_COMPLEX *a, HM_COMPLEX *roots,
    double epsilon, size_t max_iter, HM_COMPLEX wdk_init);

/* ------------------------------------------------------------------------ */
/* Alle reellen und komplexen Nullstellen eines Polynoms werden ermittelt   */
/* und im Vektor "roots" abgelegt. Die reellen Koeffizienten des Polynoms   */
/* sind in a[0], ..., a[n] zu uebergeben, deren Anzahl ist im Parameter     */
/* "dim" zu uebergeben. Hinweis: Ein Polynom mit z Koeffizienten besitzt    */
/* genau (z-1) Nullstellen. Der Ergebnisvektor "roots" hat also ein Element */
/* weniger als Vektor "a". Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (Algorithmus   */
/* konvergiert nicht), MAT_ERR_DIM (falls dim < 2 oder a[0] = 0) oder       */
/* MAT_ERR_MEMORY (freier Speicher reicht nicht aus).                       */
/* ------------------------------------------------------------------------ */
int v_droots(size_t dim, const double *a, HM_COMPLEX *roots);

/* ------------------------------------------------------------------------ */
/* Siehe v_droots(). Bei dieser Variante kann zusaetzlich die gewuenschte   */
/* Genauigkeit, die maximale Anzahl der Iterationen sowie der Startwert     */
/* fuer das verwendete Weierstrass-Durand-Kerner-Iterationsverfahren        */
/* angegeben werden. Rueckgabe: MAT_OK, MAT_ERR_ITFAIL (der Algorithmus     */
/* konvergiert nicht), MAT_ERR_DIM (falls dim < 2 oder a[0] = 0) oder       */
/* MAT_ERR_MEMORY (freier Speicher reicht nicht aus).                       */
/* ------------------------------------------------------------------------ */
int v_droots2(size_t dim, const double *a, HM_COMPLEX *roots,
    double epsilon, size_t max_iter, HM_COMPLEX wdk_init);

/* ------------------------------------------------------------------------ */
/* Die FFT (Fast Fourier Transform, schnelle Fouriertransformation) wird    */
/* fuer den angegebenen Vektor "f" mit "n" komplexen Elementen berechnet.   */
/* Achtung: Die Elementanzahl "n" muss eine Zweierpotenz sein, sonst wird   */
/* die Funktion mit dem Fehlercode MAT_ERR_DIM abgebrochen. Das Ergebnis    */
/* der FFT wird wiederum im Vektor "f" zurueckgegeben. Die Elemente des     */
/* Vektors "f" werden also durch den Aufruf von "v_cfft" ueberschrieben.    */
/* Moegliche Rueckgabewerte: MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_DIM.       */
/* ------------------------------------------------------------------------ */
int v_cfft(size_t n, HM_COMPLEX *f);

/* ------------------------------------------------------------------------ */
/* Die FFT (Fast Fourier Transform, schnelle Fouriertransformation) wird    */
/* fuer den angegebenen Vektor "f" mit "n" reellen Elementen berechnet.     */
/* Achtung: Die Elementanzahl "n" muss eine Zweierpotenz sein, sonst wird   */
/* die Funktion mit dem Fehlercode MAT_ERR_DIM abgebrochen. Das Ergebnis    */
/* der FFT wird in den Vektor "y" geschrieben, welcher Platz fuer "n"       */
/* komplexe Elemente des Typs HM_COMPLEX bieten muss. Rueckgabewerte der    */
/* Funktion "v_dfft" sind: MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_DIM.         */
/* ------------------------------------------------------------------------ */
int v_dfft(size_t n, const double *f, HM_COMPLEX *y);

/* ------------------------------------------------------------------------ */
/* Die IFFT (Inverse Fast Fourier Transform) wird fuer den Vektor "y" mit   */
/* "n" komplexen Elementen berechnet. Achtung: Die Elementanzahl "n" muss   */
/* eine Zweierpotenz sein, sonst wird "v_ifft" mit dem Fehler MAT_ERR_DIM   */
/* abgebrochen. Das Ergebnis der IFFT wird im Vektor "y" zurueckgegeben,    */
/* die Elemente des Vektors "y" werden also durch den Aufruf von "v_ifft"   */
/* ueberschrieben. Rueckgabewerte: MAT_OK, MAT_ERR_MEMORY oder MAT_ERR_DIM. */
/* ------------------------------------------------------------------------ */
int v_ifft(size_t n, HM_COMPLEX *y);

/* ------------------------------------------------------------------------ */
/* Das lineare Gleichungssystem (QR) * x = y wird geloest. Vor dem Aufruf   */
/* von "lin_solve_qr" muss zunaechst eine QR-Zerlegung der Matrix erfolgen, */
/* zum Beispiel mittels "m_qr_decomp". (Alternativ kann auch die Funktion   */
/* "lin_solve" aufgerufen werden, wo beide Schritte - QR-Zerlegung und die  */
/* die Loesung des Gleichungssystems - gemeinsam durchgefuehrt werden.)     */
/* Der Rueckgabewert betraegt MAT_OK, falls das Gleichungssystem geloest    */
/* werden konnte, ansonsten MAT_ERR_NOINV.                                  */
/* ------------------------------------------------------------------------ */
int lin_solve_qr(size_t dim, double *x,
    const void *q, const void *r, const double *y);

/* ------------------------------------------------------------------------ */
/* Das lineare Gleichungssystem A * x = y wird geloest. Es wird zunaechst   */
/* eine QR-Zerlegung der Matrix durchgefuehrt und anschliessend die Loesung */
/* des Gleichungssystems durch Rueckwaerts-Einsetzen bestimmt. Moegliche    */
/* Rueckgabewerte sind MAT_OK, MAT_ERR_MEMORY und MAT_ERR_NOINV. Bitte die  */
/* Kommentare zur Funktion "lin_solve_qr" ebenfalls beachten.               */
/* ------------------------------------------------------------------------ */
int lin_solve(size_t dim, double *x, const void *a, const double *y);

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird im CSV-Format in die bereits geoeffnete Datei      */
/* "file" geschrieben. Die Datei bleibt nach dem Schreiben geoeffnet.       */
/* Rueckgabewert == MAT_OK, falls die Matrix erfolgreich ausgegeben wurde   */
/* oder MAT_ERR_FWRITE, falls ein Fehler bei der Ausgabe aufgetreten ist.   */
/* ------------------------------------------------------------------------ */
int m_writecsv2(size_t dim, const void *mat, FILE *file);

/* ------------------------------------------------------------------------ */
/* Es wird die Datei "filename" erstellt und die Matrix "mat" im CSV-Format */
/* gespeichert. Rueckgabewert == MAT_OK, falls die Ausgabe erfolgreich war, */
/* MAT_ERR_FOPEN bzw. MAT_ERR_FWRITE, falls ein Fehler aufgetreten ist.     */
/* ------------------------------------------------------------------------ */
int m_writecsv(size_t dim, const void *mat, const char *filename);

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird aus der bereits geoeffneten CSV-Datei "file"       */
/* eingelesen, danach bleibt die Datei geoeffnet. Falls die Matrix korrekt  */
/* eingelesen wurde, ist der Rueckgabewert == MAT_OK, sonst MAT_ERR_FORMAT. */
/* ------------------------------------------------------------------------ */
int m_readcsv2(size_t dim, void *mat, FILE *file);

/* ------------------------------------------------------------------------ */
/* Die Matrix "mat" wird aus der CSV-Datei "filename" eingelesen. Falls     */
/* die Matrix korrekt eingelesen wurde, ist der Rueckgabewert == MAT_OK,    */
/* ansonsten MAT_ERR_FORMAT oder MAT_ERR_FOPEN.                             */
/* ------------------------------------------------------------------------ */
int m_readcsv(size_t dim, void *mat, const char *filename);

#ifdef __cplusplus
}
#endif

#endif  /* MATRIX_H_Included */
